"""This addon is a bridge between Blender and Ragdoll"""

import os
import sys
import platform

import bpy

bl_info = {
    "name": "Ragdoll",
    "version": (2024, 2, 5),      # Auto generated. DO NOT TOUCH!
    "blender": (3, 4, 0),         # The minimum Blender version required
    "ragdollcore": (2024, 2, 5),  # The minimum Ragdoll Core version required
    "category": "Animation",
    "author": "Ragdoll Dynamics",
    "description": "Animate with real-time physics.",
    "location": "Physics Properties",
    "doc_url": "https://learn.ragdolldynamics.com/overview/",
    "tracker_url": "https://forums.ragdolldynamics.com/",
}


platform_name = platform.system().lower()
core_version = ".".join("%02d" % i for i in bl_info["ragdollcore"])


def get_install_dir():
    if platform_name == "windows":
        root = os.getenv("PROGRAMFILES")

    elif platform_name == "darwin":
        root = "Not yet implemented"

    elif platform_name == "linux":
        root = "Not yet implemented"

    else:
        # Cannot raise exception from init
        root = "Unsupported OS: %s" % platform_name

    path = os.path.join(root, "Ragdoll Dynamics", "Core", core_version, "lib")

    # Enable override via environment variable
    path = os.getenv("RAGDOLL_CORE_PATH", path)

    return path


# Expose .pyd extension
root = get_install_dir()

# Enable the user to report where Core was attempted
print("Looking for Ragdoll Core @ '%s'" % root)
dll = os.path.join(root, "%sragdollcore.%s" % ("", "dll"))

if bpy.app.version < (3, 4, 0):
    """Handle incompatible Blender version

    We only support Blender 3.4 and above

    """

    def register():
        from . import placeholder
        placeholder.install("old_version", "Requires Blender 3.4 or above",)

    def unregister():
        from . import placeholder
        placeholder.uninstall("old_version")


elif platform_name == "darwin" and bpy.app.version >= (3, 5, 0):
    """Handle unsupported Metal renderer

    On MacOS, Blender 3.5+ does not use OpenGL, but we do.

    """

    def register():
        from . import placeholder
        placeholder.install("metal", "On MacOS, Blender 3.4 is required",)

    def unregister():
        from . import placeholder
        placeholder.uninstall("metal")

elif not os.path.exists(dll):
    """Handle missing Ragdoll Core

    The user will need Ragdoll Core installed in
    order to use this Blender Addon

    """

    if platform_name == "windows":
        def download_page(*args):
            url = "https://learn.ragdolldynamics.com/download"

            # Help those looking at the terminal
            print("Go to %s" % url)

            # Help those looking at their UI with a browser available
            import webbrowser
            webbrowser.open_new(url)

        def register():
            from . import placeholder
            placeholder.install(
                "missing_core",
                "Download Ragdoll Core %s" % core_version,
                download_page
            )

        def unregister():
            from . import placeholder
            placeholder.uninstall("missing_core")

    elif platform_name == "darwin":
        def register():
            from . import placeholder
            placeholder.install("macos", "Coming soon for MacOS")

        def unregister():
            from . import placeholder
            placeholder.uninstall("macos")

    else:
        def register():
            from . import placeholder
            placeholder.install("macos", "Coming soon for Linux")

        def unregister():
            from . import placeholder
            placeholder.uninstall("macos")


else:
    """Handle the supported case

    All is well, let's go!

    """

    # Expose ragdollcore.pyd
    sys.path.insert(0, root)

    def register():
        from . import main
        main.install()

    def unregister():
        from . import main
        main.uninstall()

        # Ensure ragdollc is uninstalled, for release of lease
        import ragdollc
        ragdollc.uninstall()
