import bpy

import ragdollc
from ragdollc import registry

from .. import scene, types
from ..vendor import bpx


@scene.with_properties("rdMarker.json")
class RdMarkerPropertyGroup(scene.PropertyGroup):
    type = "rdMarker"

    @classmethod
    def on_property_changed(cls, entity, name):
        xobj = bpx.alias(entity)

        if not xobj:
            return

        if name == "parentMarker":
            parent = registry.get("ParentComponent", entity)
            prop = xobj[name]
            xparent = prop.read()

            if xparent:
                xparent = bpx.BpxType(xparent)
                parent.entity = xparent.data["entity"]
            else:
                parent.entity = registry.null

            registry.replace(entity, parent)

        if name == "density":
            rigid = registry.get("RigidComponent", entity)
            rigid.densityCustom = xobj[name].read()

        if name == "originMatrix":
            origin = registry.get("OriginComponent", entity)
            origin.value = types.to_rdtype(xobj[name].read())

        if name in ("collide",
                    "friction",
                    "restitution",
                    "thickness",
                    "ignoreGravity",
                    "ignoreFields",
                    "wakeCounter",
                    "positionIterations",
                    "velocityIterations",
                    "maxContactImpulse",
                    "maxDepenetrationVelocity"):

            rigid = registry.get("RigidComponent", entity)
            setattr(rigid, name, types.to_rdtype(xobj[name].read()))

        if name in ("shapeType",
                    "shapeOffset",
                    "shapeRotation",
                    "shapeLength",
                    "shapeExtents",
                    "shapeVertexLimit",
                    "shapeRadius"):

            attr = name[5:]  # shapeType -> Type
            attr = attr[0].lower() + attr[1:]  # Type -> type
            desc = registry.get("GeometryDescriptionComponent", entity)
            setattr(desc, attr, types.to_rdtype(xobj[name].read()))

        if name == "inputGeometryMatrix":
            markerUi = registry.get("MarkerUIComponent", entity)
            mtx = types.to_rdmatrix(xobj["inputGeometryMatrix"].read())
            markerUi.inputGeometryMatrix = mtx

        super().on_property_changed(entity, name)


def touch_all_properties(entity):
    RdMarkerPropertyGroup.touch_all_properties(entity)


def post_constructor(xobj):
    entity = xobj.data.get("entity")

    if entity is not None:
        ragdollc.registry.destroy(entity)

    entity = ragdollc.scene.createMarker(xobj.name())

    # Create a two-way mapping between these
    bpx.create_alias(entity, xobj)

    xobj.data["entity"] = entity

    touch_all_properties(entity)


def scale_from_matrix(mtx):
    return bpx.Vector((
        mtx.col[0].length,
        mtx.col[1].length,
        mtx.col[2].length
    ))


@bpx.with_cumulative_timing
def evaluate_start_state(entity):
    xobj = bpx.alias(entity)

    removed = registry.get("RemovedComponent", entity)
    removed.value = not xobj.is_alive()

    if removed.value:
        return

    xsource = xobj["sourceTransform"].read()
    removed.value = not xsource

    if removed.value:
        return

    # Keep this up to date
    xsource.data["entity"] = entity

    mat = xsource.matrix()
    rest = registry.get("RestComponent", entity)
    rest.value = types.to_rdmatrix(mat)

    scale = registry.get("ScaleComponent", entity)
    scale.value = types.to_rdvector(scale_from_matrix(mat))

    markerUi = registry.get("MarkerUIComponent", entity)
    markerUi.sourceTransform = xsource.name()
    markerUi.limitStiffness = xobj["limitStiffness"].read()
    markerUi.limitDampingRatio = xobj["limitDampingRatio"].read()
    markerUi.collisionGroup = xobj["collisionGroup"].read()
    markerUi.drawLimit = xobj["drawLimit"].read()
    markerUi.drawDrive = xobj["drawDrive"].read()
    markerUi.mass = xobj["mass"].read()
    markerUi.centerOfMass = types.to_rdvector(xobj["centerOfMass"].read())
    markerUi.angularMass = types.to_rdvector(xobj["angularMass"].read())
    markerUi.recordTranslation = xobj["recordTranslation"].read()
    markerUi.recordRotation = xobj["recordRotation"].read()

    color = registry.get("ColorComponent", entity)
    color.value = types.to_rdcolor(xobj["color"].read())

    subs = registry.get("SubEntitiesComponent", entity)
    if registry.get("JointChangedComponent", subs.relative).value:
        registry.get("JointChangedComponent", subs.relative).value = False

        joint = registry.get("JointComponent", subs.relative)
        joint.ignoreMass = xobj["ignoreMass"].read()
        joint.child = entity

        markerUi.parentFrame = types.to_rdmatrix(xobj["parentFrame"].read())
        markerUi.childFrame = types.to_rdmatrix(xobj["childFrame"].read())

        parent = registry.get("ParentComponent", entity)
        limit = registry.get("LimitComponent", subs.relative)
        limit.enabled = False

        if registry.valid(parent.entity):
            joint.parent = parent.entity

            range_ = xobj["limitRange"].read()
            limit.twist = range_.x
            limit.swing1 = range_.y
            limit.swing2 = range_.z
            limit.enabled = True

    drawable = registry.get("DrawableComponent", entity)
    drawable.displayType = xobj["displayType"].read()
    drawable.scale = xobj["drawScale"].read()

    limitDrawable = registry.get("LimitDrawableComponent", subs.relative)
    limitDrawable.visible = markerUi.drawLimit

    drive = registry.get("DriveComponent", subs.relative)
    drive.slerp = xobj["driveInterpolation"].read()
    drive.linearAmount = types.to_rdvector(xobj["driveLinearAmount"].read())
    drive.angularAmountTwist = xobj["driveAngularAmountTwist"].read()
    drive.angularAmountSwing = xobj["driveAngularAmountSwing"].read()

    evaluate_current_state(entity)


@bpx.with_cumulative_timing
def evaluate_current_state(entity):
    xobj = bpx.alias(entity)

    # Cannot unremove an object past the start frame
    if registry.get("RemovedComponent", entity).value:
        return

    xsource = xobj["sourceTransform"].read()
    mat = xsource.matrix()

    kinematic = registry.get("KinematicComponent", entity)
    kinematic.value = types.to_rdmatrix(mat)

    markerUi = registry.get("MarkerUIComponent", entity)
    markerUi.inputType = xobj["inputType"].read()
    markerUi.linearMotion = xobj["linearMotion"].read()
    markerUi.airDensity = xobj["airDensity"].read()
    markerUi.linearStiffness = xobj["linearStiffness"].read()
    markerUi.linearDampingRatio = xobj["linearDampingRatio"].read()
    markerUi.angularStiffness = xobj["angularStiffness"].read()
    markerUi.angularDampingRatio = xobj["angularDampingRatio"].read()

    rigid = registry.get("RigidComponent", entity)
    rigid.linearDamping = markerUi.airDensity * xobj["linearDamping"].read()
    rigid.angularDamping = markerUi.airDensity * xobj["angularDamping"].read()

    subs = registry.get("SubEntitiesComponent", entity)
    joint_anim = registry.get("JointAnimation", subs.relative)
    joint_anim.influence = xobj["influence"].read()

    drive = registry.get("DriveComponent", subs.relative)
    drive.target = types.to_rdmatrix(mat)
    drive.slerp = xobj["driveInterpolation"].read()
    drive.linearAmount = types.to_rdvector(xobj["driveLinearAmount"].read())
    drive.acceleration = xobj["driveSpringType"].read()
    drive.angularAmountTwist = xobj["driveAngularAmountTwist"].read()
    drive.angularAmountSwing = xobj["driveAngularAmountSwing"].read()


def parse_input_geometry(entity):
    xobj = bpx.alias(entity)

    inputMesh = registry.get("InputMeshComponent", entity)
    inputMesh.vertices.clear()
    inputMesh.triangleIndices.clear()

    geo = xobj["inputGeometry"].read()

    # There may not be any geometry connected
    if not geo:
        return

    handle = geo.handle()
    dg = bpy.context.evaluated_depsgraph_get()

    try:
        mesh = handle.evaluated_get(dg).to_mesh()
    except RuntimeError:
        # No mesh here, this would mean the connected
        # object is not a mesh
        return

    for v in mesh.vertices:
        point = types.to_rdpoint(v.co)
        inputMesh.vertices.append(point)

    if len(inputMesh.vertices) == 0:
        return False

    for loops in mesh.loop_triangles:
        for index in loops.vertices:
            inputMesh.triangleIndices.append(index)


def install():
    scene.post_constructors["rdMarker"] = post_constructor
    scene.register_property_group(RdMarkerPropertyGroup)


def uninstall():
    scene.unregister_property_group(RdMarkerPropertyGroup)
