"""Scripting interface for Ragdoll operators

Operators use this module to perform scene manipulation and anything that
does not involve user interactivity or selection.

This module shall not access any operators, and is intended for either
operators or scripting of Ragdoll via Python.

"""

import bpy
import ragdollc
from ragdollc import registry

from .vendor import bpx
from . import scene, util


def reassign(transform, marker):
    assert isinstance(transform, bpx.BpxType), (
        "%s was not a BpxObject or BpxBone" % transform
    )
    assert marker and marker.type() == "rdMarker", (
        "%s was not a rdMarker" % marker
    )

    entry = {
        "object": transform.handle(),
    }

    if isinstance(transform, bpx.BpxBone):
        entry.update({
            "boneid": transform.boneid()
        })

    marker["sourceTransform"] = entry


def retarget(transforms, marker, append=False):
    """Retarget `marker` to `transform`

    When recording, write simulation from `marker` onto `transform`,
    regardless of where it is assigned.

    Arguments:
        marker (bpx.BpxType): Marker object.
        transforms (list[bpx.BpxBone | bpx.BpxObject]): An object or pose-bone to
            bake simulation to.
        append: Append one more transform into recording list if `True`.
            Previous targets will be replaced by `transform` if `False`.

    """

    assert marker and marker.type() == "rdMarker", (
        "%s was not a rdMarker" % marker
    )

    if not append:
        untarget(marker)

    for transform in transforms:
        entry = {
            "object": transform.handle(),
        }

        if isinstance(transform, bpx.BpxBone):
            entry.update({
                "boneid": transform.boneid()
            })

        marker["destinationTransforms"].append(entry)


def untarget(marker):
    """Remove all recording targets from `marker`

    Arguments:
        marker (bpx.BpxType): Marker object.

    """

    assert marker and marker.type() == "rdMarker", (
        "%s was not a rdMarker" % marker
    )

    count = 0

    # Remove `entity` from old target
    for dest in marker["destinationTransforms"]:
        xobj = scene.source_to_object(dest)
        xobj.data.pop("entity", None)
        count += 1

    marker["destinationTransforms"].clear()

    return count


def create_group(solver, name="rGroup"):
    group = scene.create("rdGroup", name=name)
    bpx.link(group, util.find_assembly())

    solver["members"].append({"object": group.handle()})
    return group


def move_to_group(markers, group):
    assert all(isinstance(m, bpx.BpxType) for m in markers), (
        "%s was not a series of markers"
    )
    assert isinstance(group, bpx.BpxObject) and group.type() == "rdGroup", (
        "%s was not a rdGroup"
    )

    for marker in markers:
        group["members"].append({"object": marker.handle()})

    # Dirty "members" property
    for xobj in bpx.ls(type="rdSolver"):
        entity = xobj.data.get("entity")
        ragdollc.scene.propertyChanged(entity, "members")


def remove_from_group(markers, group=None):
    """Remove `markers` from `group` or all groups"""
    if group is None:
        groups = bpx.ls(type="rdGroup")
    else:
        groups = (group,)

    for group in groups:
        for marker in markers:
            try:
                index = group["members"].index({"object": marker.handle()})
            except IndexError:
                continue

            # Ungroup
            group["members"].remove(index)

    # Dirty "members" property
    for xobj in bpx.ls(type="rdSolver"):
        entity = xobj.data.get("entity")
        ragdollc.scene.propertyChanged(entity, "members")


def cache(solver):
    for _ in cache_iter([solver]):
        pass  # Iter from start frame to end


def uncache(solver):
    solver["cache"] = False
    ragdollc.scene.evaluate(solver.data["entity"])


def cache_iter(solvers):
    """Persistently store the simulated result of the `solvers`

    Use this to scrub the timeline both backwards and forwards without
    re-simulating anything.

    """
    context_scene = bpy.context.scene

    # Remember where we came from
    initial_frame = context_scene.frame_current

    entities = {s.data["entity"] for s in solvers}
    start_frame = min({
        registry.get("TimeComponent", solver_entity).startFrame
        for solver_entity in entities
    })
    start_frame = int(start_frame)
    end_frame = context_scene.frame_end
    total = end_frame - start_frame

    context_scene.frame_set(start_frame)
    for solver in solvers:
        # Clear existing cache
        solver["cache"] = False
        ragdollc.scene.evaluate(solver.data["entity"])

        # Updated cache
        solver["cache"] = True

    for frame in range(start_frame, end_frame + 1):
        context_scene.frame_set(frame)

        for solver_entity in entities:
            ragdollc.scene.evaluate(solver_entity)

        percentage = 100 * float(frame - start_frame) / total
        yield percentage

    # Restore where we came from
    context_scene.frame_set(initial_frame)
