import bpy

from ragdollc import registry

from . import tag_redraw, eyedropper, get_selected, find_transform, find_marker
from .. import constants, commands
from ..vendor import bpx


def _get_scene_entity(xmember):
    member_entity = xmember.data["entity"]
    Scene = registry.get("SceneComponent", member_entity)
    return Scene.entity


def _has_same_solver(selection):
    solvers = set(_get_scene_entity(x) for x in selection)
    return len(solvers) == 1


class Reparent(bpy.types.Operator):
    bl_idname = "ragdoll.reparent"
    bl_label = "Reparent"
    bl_options = {"REGISTER", "UNDO"}
    bl_description = "Change the parent of the selected marker"

    @classmethod
    def poll(cls, context):
        selection = get_selected("rdMarker")

        if len(selection) == 2:
            if _has_same_solver(selection):
                return True

            else:
                cls.poll_message_set("Selected markers are not from the "
                                     "same solver.")
                return False
        else:
            cls.poll_message_set("Requires 2 markers selected, "
                                 "one child and then one parent.")
            return False

    def execute(self, context):
        child, parent = get_selected("rdMarker")
        child["parentMarker"] = parent.handle()

        self.report({"INFO"}, "Reparented %s -> %s" % (child, parent))
        return {"FINISHED"}


class Retarget(bpy.types.Operator):
    bl_idname = "ragdoll.retarget"
    bl_label = "Retarget"
    bl_options = {"REGISTER", "UNDO"}
    bl_description = "Set recording targets to the selected marker"

    @classmethod
    def poll(cls, context):
        selection = bpx.selection()

        if len(selection) < 1:
            cls.poll_message_set("Select one marker and one target transform")
            return False

        return True

    def execute(self, context):
        sel = bpx.selection()

        marker = find_marker(sel)
        transform = find_transform(sel)

        if not (marker and transform):
            self.report(
                {"WARNING"},
                "Select one Marker and one target transform",
            )
            return {"CANCELLED"}

        commands.retarget([transform], marker, append=False)

        self.report({"INFO"}, "Retargeted %s -> %s" % (marker, transform))

        return {"FINISHED"}


class Untarget(bpy.types.Operator):
    bl_idname = "ragdoll.untarget"
    bl_label = "Untarget"
    bl_options = {"REGISTER", "UNDO"}
    bl_description = "Remove all targets from the selected marker"

    @classmethod
    def poll(cls, context):
        selection = bpx.selection()

        if len(selection) < 1:
            cls.poll_message_set("Select one or more markers to untarget")
            return False

        return True

    def execute(self, context):
        count = 0

        for xobj in get_selected("rdMarker"):
            count += commands.untarget(xobj)

        if count > 0:
            self.report({"INFO"}, "Untargeted %d targets" % count)
        else:
            self.report({"WARNING"}, "No targets found")

        return {"FINISHED"}


class Reassign(bpy.types.Operator):
    bl_idname = "ragdoll.reassign"
    bl_label = "Reassign"
    bl_options = {"REGISTER", "UNDO"}
    bl_description = (
        "Reassign a Marker, maintaining its mesh "
        "and destination transform(s)"
    )

    @classmethod
    def poll(cls, context):
        selection = bpx.selection()

        if len(selection) != 2:
            cls.poll_message_set("Select one Marker and one new assignee")
            return False

        return True

    def execute(self, context):
        sel = bpx.selection()

        marker = find_marker(sel)
        transform = find_transform(sel)

        if not (marker and transform):
            self.report({"WARNING"}, "Select one Marker and one new assignee")
            return {"CANCELLED"}

        commands.reassign(transform, marker)

        self.report({"INFO"}, "Reassigned %s -> %s" % (marker, transform))

        return {"FINISHED"}


class Unparent(bpy.types.Operator):
    bl_idname = "ragdoll.unparent"
    bl_label = "Unparent"
    bl_options = {"REGISTER", "UNDO"}
    bl_description = "Remove the parent from the selected marker"

    @classmethod
    def poll(cls, context):
        selection = get_selected("rdMarker")

        if len(selection) == 1:
            return True

        cls.poll_message_set("Requires exactly 1 marker to be selected.")
        return False

    def execute(self, context):
        child, = get_selected("rdMarker")
        child["parentMarker"] = None
        return {"FINISHED"}


class Group(bpy.types.Operator):
    bl_idname = "ragdoll.group"
    bl_label = "Group"
    bl_options = {"REGISTER", "UNDO"}
    bl_description = "Group the selected markers"

    @classmethod
    def poll(cls, context):
        selection = get_selected("rdMarker")

        if len(selection) > 0:
            if _has_same_solver(selection):
                return True

            else:
                cls.poll_message_set(
                    "Selected markers are not from the "
                    "same solver."
                )
                return False

        else:
            cls.poll_message_set("Requires at least 1 marker selected.")
            return False

    def execute(self, context):
        selection = get_selected("rdMarker")
        commands.remove_from_group(selection)

        solver = bpx.alias(_get_scene_entity(selection[0]))
        group = commands.create_group(solver, name="rGroup")

        # Add selected markers to group
        commands.move_to_group(selection, group)

        tag_redraw(context.screen)
        bpx.set_active(group)

        return {"FINISHED"}


class Ungroup(bpy.types.Operator):
    bl_idname = "ragdoll.ungroup"
    bl_label = "Ungroup"
    bl_options = {"REGISTER", "UNDO"}
    bl_description = "Ungroup the selected markers"

    @classmethod
    def poll(cls, context):
        selection = get_selected("rdMarker")

        if len(selection) > 0:
            if _has_same_solver(selection):
                return True

            else:
                cls.poll_message_set("Selected markers are not from the "
                                     "same solver.")
                return False

        else:
            cls.poll_message_set("Requires at least 1 marker selected.")
            return False

    def execute(self, context):
        commands.remove_from_group(get_selected("rdMarker"))
        tag_redraw(context.screen)
        return {"FINISHED"}


class MoveToGroup(bpy.types.Operator):
    bl_idname = "ragdoll.move_to_group"
    bl_label = "Move to Group"
    bl_options = {"REGISTER", "UNDO"}
    bl_description = "Move selected markers to group"

    @classmethod
    def poll(cls, context):
        markers = get_selected("rdMarker")
        groups = get_selected("rdGroup")

        if len(markers) > 0 and len(groups) == 1:
            if _has_same_solver(markers + groups):
                return True

            cls.poll_message_set("Selected markers/group are not from the "
                                 "same solver.")
            return False

        cls.poll_message_set("Requires at least 1 marker, and only 1 group "
                             "selected.")
        return False

    def execute(self, context):
        markers = get_selected("rdMarker")
        group = get_selected("rdGroup")[0]

        commands.remove_from_group(markers)
        commands.move_to_group(markers, group)

        tag_redraw(context.screen)

        return {"FINISHED"}


class ReplaceMesh(bpy.types.Operator):
    bl_idname = "ragdoll.replace_mesh"
    bl_label = "Replace Mesh"
    bl_options = {"REGISTER", "UNDO"}
    bl_description = ("Replace input of the 'Mesh' shape type with another "
                      "polygonal mesh")

    @classmethod
    def poll(cls, context):
        if not get_selected("rdMarker"):
            cls.poll_message_set("Requires at least 1 markers selected.")
            return False

        return True

    def execute(self, _context):
        # It seems like, Blender freaks out when repeating a modal operator.
        # However, if that modal operator was invoked by another regular
        # operator, Blender is fine.
        # So to support "Repeat Last", here we are.
        bpy.ops.ragdoll._replace_mesh_internal("INVOKE_DEFAULT")
        return {"FINISHED"}


class ReplaceMeshInternal(eyedropper.EyedropperOp):
    bl_idname = "ragdoll._replace_mesh_internal"
    bl_label = ReplaceMesh.bl_label
    bl_options = {"INTERNAL"}
    bl_description = ReplaceMesh.bl_description

    def hit_types(self) -> set[str]:
        return {"MESH"}

    def on_clicked(self, hit_object, context):
        selection = get_selected("rdMarker")
        if not len(selection):
            self.report({"ERROR"}, "No marker selected.")
            return {"CANCELLED"}

        maintain_offset = True

        mesh = hit_object
        if not (isinstance(mesh, bpy.types.Object) and mesh.type == "MESH"):
            self.report({"ERROR"}, "No mesh picked, this is a bug.")
            return {"CANCELLED"}

        for marker in selection:
            marker["shapeType"] = constants.MeshShape

            if maintain_offset:
                source = marker["sourceTransform"].read()
                source_mtx = source.matrix()
                mesh_mtx = mesh.matrix_world
                mesh_mtx = source_mtx.inverted_safe() @ mesh_mtx
                marker["inputGeometryMatrix"] = mesh_mtx

            marker["inputGeometry"] = {"object": mesh}

        return {"FINISHED"}


def install():
    bpy.utils.register_class(Reparent)
    bpy.utils.register_class(Reassign)
    bpy.utils.register_class(Unparent)
    bpy.utils.register_class(Retarget)
    bpy.utils.register_class(Untarget)
    bpy.utils.register_class(Group)
    bpy.utils.register_class(Ungroup)
    bpy.utils.register_class(MoveToGroup)
    bpy.utils.register_class(ReplaceMeshInternal)
    bpy.utils.register_class(ReplaceMesh)


def uninstall():
    bpy.utils.unregister_class(Reparent)
    bpy.utils.unregister_class(Reassign)
    bpy.utils.unregister_class(Unparent)
    bpy.utils.unregister_class(Retarget)
    bpy.utils.unregister_class(Untarget)
    bpy.utils.unregister_class(Group)
    bpy.utils.unregister_class(Ungroup)
    bpy.utils.unregister_class(MoveToGroup)
    bpy.utils.unregister_class(ReplaceMesh)
    bpy.utils.unregister_class(ReplaceMeshInternal)
