import os
import bpy

import ragdollc

from .. import bl_info
from . import icons

from ..operators import (
    create_pin_constraint,
    create_attach_constraint,
    record_simulation,
    assign_markers,
    delete_physics,
    edit_solver,
    edit_marker,
    retarget_ui,
    parser,
)


def _with_ctrl(layout, cls, **kwargs):
    split = layout.split(align=True, factor=0.99)

    cls.bl_description += "\n\nTip: Ctrl + Click = Options Dialog"
    op = split.operator(cls.bl_idname, **kwargs)

    split.separator(factor=0)  # Magic element for maintaining this layout.
    split.label(text="", icon_value=icons.name_to_icon_id("ctrl"))

    return op


class RagdollMainMenu(bpy.types.Menu):
    bl_label = "Ragdoll"
    bl_idname = "ANIMATION_MT_ragdoll_menu"

    @classmethod
    def poll(cls, context):
        return context.mode in {"OBJECT", "POSE"}

    def draw(self, context):
        layout = self.layout
        col = layout.column()

        col.label(text="Markers")

        if context.mode == "POSE":
            _with_ctrl(
                col,
                assign_markers.AssignMarkers,
                text="Assign and Connect",
                icon="REC",
            )
        else:
            _with_ctrl(
                col,
                assign_markers.AssignMarkers,
                text="Assign",
                icon="REC",
            )

        _soon(col, "Assign Environment", "MESH_GRID")

        col.separator(factor=0.5)
        col.label(text="Transfer")

        _with_ctrl(
            col,
            record_simulation.RecordSimulation,
            icon_value=icons.name_to_icon_id("record"),
        )
        _soon(col, "Snap to Simulation", "PROP_ON")

        col.separator(factor=0.5)
        col.label(text="IO")

        col.operator(parser.ExportPhysics.bl_idname,
                     icon="EXPORT")
        _soon(col, "Import Physics", "CON_ARMATURE")
        col.operator(parser.LoadPhysics.bl_idname,
                     icon="IMPORT")
        _soon(col, "Update Physics", "CON_ARMATURE")

        col.separator(factor=0.5)
        col.label(text="Manipulate")

        col.menu(RagdollConstrainMenu.bl_idname, icon="LINKED")
        col.menu(RagdollEditMenu.bl_idname, icon="ORIENTATION_GLOBAL")
        col.menu(RagdollFieldsMenu.bl_idname, icon="FORCE_WIND")

        col.separator(factor=0.5)
        col.label(text="Utilities")

        col.menu(RagdollUtilitiesMenu.bl_idname, icon="SNAP_ON")
        col.menu(RagdollSystemMenu.bl_idname, icon="DESKTOP")

        col.separator(factor=0.5)
        col.menu(RagdollAssetsMenu.bl_idname, icon="MESH_MONKEY")

        col.operator("ragdoll.licence",
                     text="Ragdoll %s" % ragdollc.__version__,
                     icon_value=icons.name_to_icon_id("logo"))


class RagdollUtilitiesMenu(bpy.types.Menu):
    bl_label = "Utilities"
    bl_idname = "ANIMATION_MT_ragdoll_menu_utilities"

    def draw(self, _context):
        layout = self.layout
        col = layout.column()

        col.operator(edit_marker.ReplaceMesh.bl_idname,
                     icon="MOD_PARTICLE_INSTANCE")

        col.separator(factor=0.5)
        col.operator(record_simulation.ExtractSimulation.bl_idname,
                     icon="MOD_ARMATURE")

        col.separator(factor=0.5)
        # TODO: Utilities
        _soon(col, "Auto Limit", "ORIENTATION_LOCAL")
        _soon(col, "Reset Shape", "MESH_CAPSULE")
        _soon(col, "Reset Origin", "PIVOT_BOUNDBOX")
        _soon(col, "Reset Constraint Frames", "RIGID_BODY_CONSTRAINT")
        _soon(col, "Edit Constraint Frames", "MOD_MESHDEFORM")


class RagdollSystemMenu(bpy.types.Menu):
    bl_label = "System"
    bl_idname = "ANIMATION_MT_ragdoll_menu_system"

    def draw(self, _context):
        layout = self.layout
        col = layout.column()

        col.operator(delete_physics.DeletePhysicsAll.bl_idname,
                     icon="GHOST_DISABLED")

        col.operator(delete_physics.DeletePhysicsBySelection.bl_idname,
                     icon="GHOST_DISABLED")

        col.separator(factor=0.5)
        col.menu(RagdollLoggingMenu.bl_idname, icon="INFO")


class RagdollConstrainMenu(bpy.types.Menu):
    bl_label = "Constrain"
    bl_idname = "ANIMATION_MT_ragdoll_menu_constrain"

    def draw(self, _context):
        layout = self.layout
        col = layout.column()

        _soon(col, "Distance", "DRIVER_DISTANCE")

        _with_ctrl(
            col, create_pin_constraint.CreatePinConstraint,
            text="Pin",
            icon="GP_ONLY_SELECTED"
        )

        _with_ctrl(
            col, create_attach_constraint.CreateAttachConstraint,
            text="Attach",
            icon="PIVOT_MEDIAN"
        )

        _soon(col, "Weld", "ORIENTATION_LOCAL")


class RagdollEditMenu(bpy.types.Menu):
    bl_label = "Edit"
    bl_idname = "ANIMATION_MT_ragdoll_menu_edit"

    def draw(self, _context):
        layout = self.layout
        col = layout.column()

        col.label(text="Hierarchy")

        col.operator(edit_marker.Reassign.bl_idname,
                     icon="MOD_PARTICLES")

        _with_ctrl(
            col,
            retarget_ui.RetargetMenuOp,
            text="Retarget",
            icon="CON_TRACKTO"
        )

        col.operator(edit_marker.Reparent.bl_idname,
                     icon="MOD_PARTICLES")

        col.separator(factor=0.5)

        col.operator(edit_marker.Unparent.bl_idname,
                     icon="PIVOT_ACTIVE")
        col.operator(edit_marker.Untarget.bl_idname,
                     icon="PIVOT_ACTIVE")

        col.separator(factor=0.5)

        col.label(text="Membership")

        col.operator(edit_marker.Group.bl_idname,
                     icon="PACKAGE")
        col.operator(edit_marker.Ungroup.bl_idname,
                     icon="UGLYPACKAGE")
        col.operator(edit_marker.MoveToGroup.bl_idname,
                     icon="COLLECTION_NEW")

        col.separator(factor=0.5)
        col.label(text="Collisions")

        _soon(col, "Assign Collision Group", "MOD_PHYSICS")
        _soon(col, "Add to Collision Group", "MOD_BOOLEAN")
        _soon(col, "Remove from Collision Group", "MOD_EDGESPLIT")

        col.separator(factor=0.5)
        col.label(text="Solver")

        _soon(col, "Merge Solvers", "PIVOT_INDIVIDUAL")
        _soon(col, "Extract Markers", "PIVOT_ACTIVE")
        _soon(col, "Move to Solver", "PIVOT_MEDIAN")

        col.separator(factor=0.5)
        col.label(text="Cache")

        col.operator(edit_solver.CacheAll.bl_idname, icon="NODE_SEL")
        col.operator(edit_solver.Uncache.bl_idname, icon="NODE")


class RagdollFieldsMenu(bpy.types.Menu):
    bl_label = "Fields"
    bl_idname = "ANIMATION_MT_ragdoll_menu_fields"

    def draw(self, _context):
        layout = self.layout
        col = layout.column()

        _soon(col, "Air", "FORCE_WIND")
        _soon(col, "Drag", "FORCE_DRAG")
        _soon(col, "Gravity", "LIGHTPROBE_CUBEMAP")
        _soon(col, "Newton", "SORTBYEXT")
        _soon(col, "Radial", "PROP_CON")
        _soon(col, "Turbulence", "FORCE_TURBULENCE")
        _soon(col, "Uniform", "FORCE_FORCE")
        _soon(col, "Vortex", "FORCE_VORTEX")
        _soon(col, "Volume Axis", "MESH_CUBE")
        _soon(col, "Volume Curve", "FORCE_CURVE")


class RagdollLoggingMenu(bpy.types.Menu):
    bl_label = "Logging Level"
    bl_idname = "ANIMATION_MT_ragdoll_menu_logging"

    def draw(self, _context):
        layout = self.layout
        col = layout.column()

        col.operator("ragdoll.logging_off")
        col.operator("ragdoll.logging_default")
        col.operator("ragdoll.logging_less")
        col.operator("ragdoll.logging_more")


class RagdollAssetsMenu(bpy.types.Menu):
    bl_label = "Assets"
    bl_idname = "ANIMATION_MT_ragdoll_menu_assets"

    _icons = {
        "alien": "PACKAGE",
        "batty": "EXPERIMENTAL",
        "cowboy": "FCURVE",
        "dog": "RENDER_RESULT",
        "dude": "EMPTY_DATA",
        "lion": "RENDER_ANIMATION",
        "manikin": "ARMATURE_DATA",
        "pirate": "RNA",
        "rhino": "OUTLINER_OB_LATTICE",
        "shark": "META_CAPSULE",
        "skeleton": "FORCE_LENNARDJONES",
        "spaceman": "FORCE_BOID",
        "wasp": "CON_SIZELIMIT",
        "wyvern": "DRIVER",
    }

    def draw(self, _context):
        layout = self.layout
        col = layout.column()
        col.operator_context = "EXEC_DEFAULT"

        assets = _LoadAsset.assets.copy()

        # Easier to find our famous manikin :)
        manikin = "manikin"
        op = col.operator(_LoadAsset.bl_idname,
                          text=manikin.capitalize(),
                          icon=self._icons.get(manikin))
        op.filepath = assets.pop(manikin)

        col.separator(factor=0.5)

        for name, path in assets.items():
            icon = self._icons.get(name)
            op = col.operator(_LoadAsset.bl_idname,
                              text=name.capitalize(),
                              icon=icon)
            op.filepath = path


def _soon(layout, text, icon):
    col = layout.column()
    col.operator("ragdoll._coming_soon", text=text, icon=icon)
    col.enabled = False


class _ComingSoon(bpy.types.Operator):
    bl_idname = "ragdoll._coming_soon"
    bl_label = ""
    bl_options = {"INTERNAL"}
    bl_description = "Coming soon..."

    def execute(self, _context):
        return {"FINISHED"}


class _LoadAsset(bpy.types.Operator):
    bl_idname = "ragdoll._load_asset"
    bl_label = "Load Physics Asset"
    bl_options = {"INTERNAL"}
    bl_description = "Load ragdoll physics"

    assets = dict()
    filepath: bpy.props.StringProperty()

    def execute(self, context):
        bpy.ops.ragdoll.load_physics(filepath=self.filepath)
        return {"FINISHED"}

    @classmethod
    def load_assets(cls):
        cls.assets.clear()

        dirname = os.path.dirname(os.path.dirname(__file__))  # ragdoll
        dirname = os.path.join(dirname, "resources", "assets")

        if not os.path.isdir(dirname):
            return

        for item in os.listdir(dirname):
            path = os.path.join(dirname, item)

            if item.endswith(".rag") and os.path.isfile(path):
                name, _ = os.path.splitext(item)
                cls.assets[name] = path


_classes = (
    RagdollMainMenu,
    RagdollUtilitiesMenu,
    RagdollSystemMenu,
    RagdollConstrainMenu,
    RagdollEditMenu,
    RagdollFieldsMenu,
    RagdollLoggingMenu,
    RagdollAssetsMenu,

    _ComingSoon,
    _LoadAsset
)


def _draw_menu(self, _context):
    self.layout.menu(RagdollMainMenu.bl_idname)


def install():
    _LoadAsset.load_assets()

    for cls in _classes:
        bpy.utils.register_class(cls)

    bpy.types.VIEW3D_MT_editor_menus.append(_draw_menu)


def uninstall():
    for cls in reversed(_classes):
        bpy.utils.unregister_class(cls)

    bpy.types.VIEW3D_MT_editor_menus.remove(_draw_menu)
