import bpy
import itertools

from . import draw
from .. import constants, scene
from ..vendor import bpx
from ..operators import retarget_ui


class _TargetPair(bpy.types.PropertyGroup):
    depth: bpy.props.IntProperty()
    dst_index: bpy.props.IntProperty(default=-1)
    marker_name: bpy.props.StringProperty()  # For filtering
    marker_handle: bpy.props.PointerProperty(type=bpy.types.Object)


class RdSolverUiPropertyGroup(bpy.types.PropertyGroup):
    type = "rdSolverUi"

    # TODO: This need rebuild after scene open, or rather,
    #  whenever solver membership changed or marker destination changed.
    targets: bpy.props.CollectionProperty(type=_TargetPair)

    def update_selection_from_ui_index(self, context):
        if not self.targets_sel_sync:
            return

        if context.mode != "OBJECT":
            return  # Marker can only be selected in Object Mode.

        if not context.space_data.use_pin_id:
            return  # We lost UI after selecting marker if not pinned.

        active = self.targets[self.targets_ui_index]
        if active.marker_handle:
            bpx.select([active.marker_handle.name_full])

    targets_ui_index: bpy.props.IntProperty(
        update=update_selection_from_ui_index,
        description="The index of current retargeting marker"
    )

    targets_sel_sync: bpy.props.BoolProperty(
        default=True,
        description="Sync marker selection with scene. (require 'Pin ID' "
                    "toggled)",
    )

    # The hash value of one `bpy.types.Window` instance which was opened as
    # retargeting window.
    #
    # A Window?
    # To have a complete marker retargeting workflow in Blender, a persistent
    # recording target listing view is required. Instead of adding another tab
    # into 3D-View SideBar, we open another window for this. And that window
    # is set to show Properties Panel for target list. (solver object pinned)
    #
    # Why Hash Window?
    # In Properties Panel, solver object's properties and target list are both
    # displayed. But we are only interested in target list in that retargeting
    # window. Other panels should be hidden. So we compare hash value of that
    # `context.window` against our `targets_window` to decide if panel should
    # be rendered.
    targets_window: bpy.props.StringProperty()

    def update_ui_index_from_selection(self, context):
        if not self.targets_sel_sync:
            return

        if not len(self.targets):
            return

        if context.mode != "OBJECT":
            return  # Marker can only be active in Object Mode.

        selection = bpx.selection(type="rdMarker", active=True)
        if selection:
            marker_handle = selection[0].handle()

            @bpx.deferred
            def set_ui_index(ui_index):
                _tmp = self.targets_sel_sync
                self.targets_sel_sync = False
                self.targets_ui_index = ui_index
                self.targets_sel_sync = _tmp

            active = self.targets[self.targets_ui_index]
            if active.marker_handle != marker_handle:

                for index, item in enumerate(self.targets):
                    if item.marker_handle == marker_handle:
                        set_ui_index(index)
                        break

    def targets_rebuild(self):

        self.targets.clear()

        solver = bpx.BpxType(self.id_data)

        markers = []
        for member in solver["members"]:
            xobj = bpx.BpxType(member.object)
            if xobj.type() == "rdMarker":
                markers.append(xobj)

        children_by_parent = {
            parent_marker: [
                child_marker
                for child_marker in markers
                if child_marker["parentMarker"].read() == parent_marker
            ]
            for parent_marker in itertools.chain(
                markers,
                [None],  # parent of root marker
            )
        }

        def list_children(parent_, depth_):
            return [
                (child_, depth_) for child_ in children_by_parent[parent_]
            ]

        def walk_hierarchy(markers_):  # Depth first
            root_markers = [
                (root_marker, 0)
                for root_marker in children_by_parent[None]
                if root_marker in markers_
            ]
            visiting_list = root_markers

            while visiting_list:
                visited_marker, depth_ = visiting_list[0]
                parent_marker = visited_marker["parentMarker"].read()

                _child_num = len(children_by_parent[visited_marker])
                _sibling_num = len(children_by_parent[parent_marker])
                _retain_level = _sibling_num <= 1 and _child_num <= 1
                # Advance
                next_dep = depth_ if _retain_level else (depth_ + 1)
                children = list_children(visited_marker, next_dep)
                visiting_list = children + visiting_list[1:]

                yield visited_marker, depth_

        def new_pair(marker_, depth_, index_=-1) -> _TargetPair:
            _pair = self.targets.add()
            _pair.depth = depth_
            _pair.dst_index = index_
            _pair.marker_name = marker_.name()
            _pair.marker_handle = marker_.handle()
            return _pair

        for marker, depth in walk_hierarchy(markers):
            dst_count = len(marker["destinationTransforms"].read())

            if dst_count == 0:
                new_pair(marker, depth)
                continue

            for index in range(dst_count):
                new_pair(marker, depth, index)


@draw.with_properties("rdSolver.json")
class RD_PT_Solver(draw.PropertiesPanel):
    bl_label = ""
    bl_space_type = "PROPERTIES"
    bl_region_type = "WINDOW"
    bl_context = "physics"

    @classmethod
    def poll(cls, context):
        xobj = bpx.BpxType(context.object)
        if bpx.is_type(xobj, "rdSolver"):
            current_window = str(hash(context.window))
            retarget_window = xobj.handle().rdSolverUi.targets_window
            # We do not want any panel except "Targets" if this window is
            # created for Retargeting.
            return current_window != retarget_window
        return False

    @classmethod
    def get_property_group(cls, context):
        xobj = bpx.BpxType(context.object)
        return xobj.property_group()

    def draw_header(self, _):
        draw.ragdoll_header(self.layout, text="Solver", icon="VIEW3D")

    def draw(self, context):
        pass


class RD_PT_Targets(bpy.types.Panel):
    bl_label = ""
    bl_space_type = "PROPERTIES"
    bl_region_type = "WINDOW"
    bl_context = "physics"

    @classmethod
    def poll(cls, context):
        xobj = bpx.BpxType(context.object)
        return bpx.is_type(xobj, "rdSolver")

    def draw_header(self, _context):
        draw.ragdoll_header(self.layout, text="Targets", icon="CON_TRACKTO")

    def draw(self, context):
        layout = self.layout
        solver_ui = context.object.rdSolverUi
        is_in_window = solver_ui.targets_window == str(hash(context.window))

        if is_in_window:
            row_count = retarget_ui.RetargetWindow.compute_row_count(context)
        else:
            row_count = 8

        # View
        column = layout.column()

        row = column.row()
        row.template_list(
            RD_UL_Targets.__name__,
            "",
            solver_ui,
            "targets",
            solver_ui,
            "targets_ui_index",
            rows=row_count,
            maxrows=row_count,
            type="DEFAULT",
        )

        column.separator()

        split = layout.split(factor=0.75)
        left = split.row(align=True)
        right = split.row()

        transform = retarget_ui.get_one_transform_from_selection()
        target_text = transform.name() if transform else ""

        if len(solver_ui.targets):
            active = solver_ui.targets[solver_ui.targets_ui_index]
            marker = bpx.BpxType(active.marker_handle)
            marker_entity = marker.data["entity"]
        else:
            marker_entity = 0
            split.enabled = False

        op = left.operator(
            retarget_ui.Retarget.bl_idname,
            text="Retarget to.. %s" % target_text,
            icon="CON_TRACKTO",
        )
        op.marker = marker_entity
        op.append = False

        left.menu(_RetargetUIMenu.bl_idname, icon="TRIA_DOWN")

        op = right.operator(
            retarget_ui.Untarget.bl_idname,
            text="Untarget",
            icon="RADIOBUT_OFF",
        )
        op.marker = marker_entity


class _RetargetUIMenu(bpy.types.Menu):
    bl_label = ""
    bl_idname = "ANIMATION_MT_ragdoll_menu_retarget_ui"

    def draw(self, context):
        solver_ui = context.object.rdSolverUi
        if len(solver_ui.targets):
            active = solver_ui.targets[solver_ui.targets_ui_index]
            marker = bpx.BpxType(active.marker_handle)
            marker_entity = marker.data["entity"]
        else:
            marker_entity = 0

        op = self.layout.operator(
            retarget_ui.Retarget.bl_idname,
            text="Append",
            icon="OUTLINER_OB_POINTCLOUD",
        )
        op.marker = marker_entity
        op.append = True


class RD_UL_Targets(bpy.types.UIList):
    marker_shape_icons = {
        constants.BoxShape: "MESH_CUBE",
        constants.SphereShape: "MESH_UVSPHERE",
        constants.CapsuleShape: "META_CAPSULE",
        constants.MeshShape: "MESH_ICOSPHERE",
    }

    render_indention: bpy.props.BoolProperty(
        name="Display Marker Hierarchy Indention",
        default=True,
        description="Indent marker name base on hierarchy when not ordered "
                    "by names or reversed"
    )

    def draw_item(self,
                  context,
                  layout,
                  data,
                  item: _TargetPair,
                  icon,
                  active_data,
                  active_property,
                  index=0,
                  flt_flag=0):
        # Always show filter
        self.use_filter_show = True  # noqa

        split = layout.split(factor=0.5)
        src_layout = split.row()
        dst_layout = split.row()

        solver_ui = data
        marker = bpx.BpxType(item.marker_handle)

        destinations = marker["destinationTransforms"]
        try:
            pointer = destinations[item.dst_index]  # RdPointerPropertyGroup
        except IndexError:
            dst_name = ""
            dst_is_bone = False
            dst_is_object = False
        else:
            xdst = scene.source_to_object(pointer)
            dst_name = xdst.name()
            dst_is_bone = isinstance(xdst, bpx.BpxBone)
            dst_is_object = isinstance(xdst, bpx.BpxObject)

        dst_index = item.dst_index
        dst_count = len(destinations.read())

        marker_shape = self.marker_shape_icons[marker["shapeType"].read()]
        marker_color = (*marker["color"], 1.0)

        render_indention = (self.render_indention and
                            not self.use_filter_sort_alpha and
                            not self.use_filter_sort_reverse)

        spacing = src_layout.row()
        spacing.separator()
        if render_indention:
            spacing.ui_units_x = 1.0 * item.depth

        # color dot
        subrow = src_layout.row()
        subrow.template_node_socket(color=marker_color)
        subrow.scale_x = 0.4

        # marker shape, name
        src_layout.label(text=marker.name(), icon=marker_shape)

        # selection indicator
        active = solver_ui.targets_ui_index
        indicator_icon = "BLANK1"

        if active == index:
            indicator_icon = "DOT"
        elif active > index and (active - index) < (dst_count - dst_index):
            indicator_icon = "DOT"
        elif active <= index and (index - active) <= dst_index:
            indicator_icon = "DOT"

        dst_layout.label(text="", icon=indicator_icon)

        # transform icon, name
        icon = ("BONE_DATA" if dst_is_bone else
                "MESH_DATA" if dst_is_object else
                "GHOST_DISABLED")
        dst_layout.label(text=dst_name, icon=icon)

    def draw_filter(self, context, layout):
        solver_ui = context.object.rdSolverUi
        row = layout.row()

        row.operator(
            _TargetsRebuild.bl_idname,
            text="",
            icon="FILE_REFRESH",
        )

        subrow = row.row(align=True)
        subrow.prop(
            self,
            "render_indention",
            icon="OUTLINER" if self.render_indention else "PRESET",
            toggle=1,
            icon_only=True,
        )
        _r = subrow.row(align=True)
        _r.prop(
            solver_ui,
            "targets_sel_sync",
            text="",
            icon="RESTRICT_SELECT_OFF",
            toggle=1,
            icon_only=True,
        )
        _r.enabled = (
            # Marker can only be selected in Object Mode.
            context.mode == "OBJECT" and
            # We lost UI after selecting marker if not pinned.
            context.space_data.use_pin_id
        )

        subrow = row.row(align=True)
        subrow.prop(self, "filter_name", text="")
        subrow.prop(
            self,
            "use_filter_invert",
            icon="ARROW_LEFTRIGHT",
            toggle=1,
            icon_only=True,
        )

        subrow = row.row(align=True)
        subrow.prop(
            self,
            "use_filter_sort_alpha",
            icon="SORTALPHA",
            toggle=1,
            icon_only=True,
        )
        subrow.prop(
            self,
            "use_filter_sort_reverse",
            icon="SORT_DESC" if self.use_filter_sort_reverse else "SORT_ASC",
            toggle=1,
            icon_only=True,
        )

    def filter_items(self, context, data, propname):
        solver_ui = data
        targets = getattr(solver_ui, propname)
        Helper = bpy.types.UI_UL_list

        # Default return values.
        flt_flags = []
        flt_neworder = []

        if self.filter_name:
            new_flt_flags = Helper.filter_items_by_name(
                self.filter_name,
                self.bitflag_filter_item,
                targets,
                "marker_name",
                reverse=self.use_filter_invert,
            )
            flt_flags = _merge_flt_flags(flt_flags, new_flt_flags)

        if not flt_flags:
            flt_flags = [self.bitflag_filter_item] * len(targets)

        if self.use_filter_sort_alpha:
            flt_neworder = Helper.sort_items_by_name(targets, "marker_name")

        if self.use_filter_invert:
            for idx, flag in enumerate(flt_flags):
                flt_flags[idx] = 0 if flag else self.bitflag_filter_item

        # This `filter_items` function gets called whenever scene
        # selection changed. So we use it as a callback to update
        # ui_index.
        solver_ui.update_ui_index_from_selection(context)

        return flt_flags, flt_neworder


def _merge_flt_flags(l1, l2):
    # https://github.com/blender/blender-addons/blob/main/object_collection_manager/ui.py#L1210
    for idx, _ in enumerate(l1):
        l1[idx] &= l2.pop(0)
    return l1 + l2


class _TargetsRebuild(bpy.types.Operator):
    bl_idname = "ragdoll.ui_targets_rebuild"
    bl_label = "Refresh"
    bl_options = {"INTERNAL"}
    bl_description = "Refresh recording targets"

    @classmethod
    def poll(cls, context):
        return hasattr(context.object, "rdSolverUi")

    def execute(self, context):
        solver_ui = context.object.rdSolverUi
        solver_ui.targets_rebuild()
        return {"FINISHED"}


def install():
    bpy.utils.register_class(_TargetPair)
    scene.register_property_group(RdSolverUiPropertyGroup)

    bpy.utils.register_class(RD_PT_Solver)
    bpy.utils.register_class(RD_PT_Targets)
    bpy.utils.register_class(RD_UL_Targets)

    bpy.utils.register_class(_RetargetUIMenu)
    bpy.utils.register_class(_TargetsRebuild)


def uninstall():
    bpy.utils.unregister_class(RD_PT_Solver)
    bpy.utils.unregister_class(RD_PT_Targets)
    bpy.utils.unregister_class(RD_UL_Targets)

    bpy.utils.unregister_class(_RetargetUIMenu)
    bpy.utils.unregister_class(_TargetsRebuild)

    scene.unregister_property_group(RdSolverUiPropertyGroup)
    bpy.utils.unregister_class(_TargetPair)
